@@ -46,25 +46,44 @@ import dataclasses
 # noinspection PyUnresolvedReferences,PyProtectedMember
 from argparse import (
     ArgumentParser,
-    Action,
-    _StoreAction,
-    _StoreTrueAction,
-    _StoreFalseAction,
     ArgumentTypeError,
+    _SubParsersAction,
+    Namespace,
 )
-from dataclasses import dataclass, make_dataclass, MISSING
+from dataclasses import dataclass, MISSING
 from enum import Enum
 from functools import wraps, partial
 from inspect import signature
-from typing import Callable, Dict, TypeVar, Type, Sequence, Optional
+from typing import (
+    Callable,
+    Dict,
+    TypeVar,
+    Type,
+    Sequence,
+    Optional,
+    overload,
+    Any,
+    Mapping,
+    Union,
+    cast,
+)
+
+from .compat import RecordField, RecordClass, NotARecordClass, DatargsParams
+
+
+@dataclass
+class Action:
+    args: Sequence[Any] = dataclasses.field(default_factory=list)
+    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)
 
-from .compat import RecordField, RecordClass
 
 DispatchCallback = Callable[[str, RecordField], Action]
 AddArgFunc = Callable[[RecordField], Action]
 
 
-def field_name_to_arg_name(name: str) -> str:
+def field_name_to_arg_name(name: str, positional=False) -> str:
+    if positional:
+        return name
     return f"--{name.replace('_','-')}"
 
 
@@ -90,8 +109,11 @@ class TypeDispatch:
 
 def add_name_formatting(func: DispatchCallback) -> AddArgFunc:
     @wraps(func)
-    def new_func(field):
-        return func(field_name_to_arg_name(field.name), field)
+    def new_func(field: RecordField):
+        return func(
+            field_name_to_arg_name(field.name, positional=field.is_positional),
+            field,
+        )
 
     return new_func
 
@@ -106,23 +128,22 @@ def get_option_strings(name: str, field: RecordField):
 
 
 def common_kwargs(field: RecordField):
-    return {
-        "type": field.type,
-        "dest": field.name,
-        **field.metadata,
-    }
+    return {"type": field.type, **subdict(field.metadata, ["aliases", "positional"])}
 
 
-def add_default(name, field: RecordField, cls=_StoreAction, **kwargs):
+def subdict(dct, remove_keys):
+    return {key: value for key, value in dct.items() if key not in remove_keys}
+
+
+def add_default(name, field: RecordField, **kwargs) -> Action:
     kwargs = {
-        "required": field.is_required(),
         "default": field.default,
         **common_kwargs(field),
         **kwargs,
     }
-    return call_func_with_matching_kwargs(
-        cls, get_option_strings(name, field), **kwargs
-    )
+    if not field.is_positional:
+        kwargs["required"] = field.is_required()
+    return Action(kwargs=kwargs, args=get_option_strings(name, field))
 
 
 T = TypeVar("T")
@@ -135,17 +156,21 @@ def call_func_with_matching_kwargs(func: Callable[..., T], *args, **kwargs) -> T
 
 
 @TypeDispatch.register(bool)
-def bool_arg(name: str, field: RecordField):
-    cls = (
-        _StoreFalseAction if field.default and field.has_default() else _StoreTrueAction
-    )
-    return call_func_with_matching_kwargs(
-        cls, get_option_strings(name, field), **common_kwargs(field)
+def bool_arg(name: str, field: RecordField) -> Action:
+    kwargs = {
+        **subdict(common_kwargs(field), ["type"]),
+        "action": "store_false"
+        if field.default and field.has_default()
+        else "store_true",
+    }
+    return Action(
+        args=get_option_strings(name, field),
+        kwargs=kwargs,
     )
 
 
 @TypeDispatch.register(Enum)
-def enum_arg(name: str, field: RecordField):
+def enum_arg(name: str, field: RecordField) -> Action:
     def enum_type_func(value: str):
         result = field.type.__members__.get(value)
         if not result:
@@ -166,7 +191,22 @@ def enum_arg(name: str, field: RecordField):
 ParserType = TypeVar("ParserType", bound=ArgumentParser)
 
 
-def make_parser(cls: type, parser: ParserType = None) -> ParserType:
+@overload
+def make_parser(cls: type) -> ArgumentParser:
+    pass
+
+
+@overload
+def make_parser(cls: type, parser: None = None) -> ArgumentParser:
+    pass
+
+
+@overload
+def make_parser(cls: type, parser: ParserType) -> ParserType:
+    pass
+
+
+def make_parser(cls, parser=None):
     # noinspection PyShadowingNames
     """
     Create parser that parses command-line arguments according to the fields of `cls`.
@@ -188,13 +228,78 @@ def make_parser(cls: type, parser: ParserType = None) -> ParserType:
     Namespace(first_arg=0, second_arg=1.5)
     """
     record_class = RecordClass.wrap_class(cls)
-    parser = parser or ArgumentParser()
+    return _make_parser(record_class, parser=parser)
+
+
+class DatargsSubparsers(_SubParsersAction):
+    """
+    A subparsers action that creates the correct sub-command class upon parsing.
+    """
+    def __init__(self, name, *args, **kwargs):
+        self.__name = name
+        super().__init__(*args, **kwargs)
+        self._command_type_map = {}
+
+    def add_parser(self, typ: type, name: str, *args, **kwargs):
+        result = super().add_parser(name, *args, **kwargs)
+        self._command_type_map[name] = typ
+        return result
+
+    def __call__(self, parser, namespace, values, *args, **kwargs):
+        new_ns = Namespace()
+        name, *_ = values
+        super().__call__(parser, new_ns, values)
+        setattr(namespace, self.__name, self._command_type_map[name](**vars(new_ns)))
+
+
+def _make_parser(record_class: RecordClass, parser: ParserType = None) -> ParserType:
+    if not parser:
+        parser = ArgumentParser(**record_class.parser_params)
+    assert parser is not None
     for name, field in record_class.fields_dict().items():
-        # noinspection PyProtectedMember
-        parser._add_action(TypeDispatch.add_arg(field))
+        sub_commands = None
+        try:
+            if field.type.__origin__ is Union:
+                sub_commands = field.type.__args__
+        except AttributeError:
+            pass
+        if sub_commands is not None:
+            add_subparsers(parser, record_class, field, sub_commands)
+        else:
+            action = TypeDispatch.add_arg(field)
+            parser.add_argument(*action.args, **action.kwargs)
     return parser
 
 
+def add_subparsers(
+    parser: ArgumentParser,
+    top_class: RecordClass,
+    sub_parsers_field: RecordField,
+    sub_parser_classes: Sequence[type],
+):
+    # noinspection PyArgumentList
+    subparsers = cast(
+        DatargsSubparsers,
+        parser.add_subparsers(
+            **top_class.sub_commands_params,
+            action=DatargsSubparsers,
+            name=sub_parsers_field.name,
+        ),
+    )
+    for command in sub_parser_classes:
+        try:
+            sub_parsers_args = top_class.wrap_class(command)
+        except NotARecordClass:
+            raise Exception(
+                f"{top_class.name}.{sub_parsers_field.name}: "
+                f"Union must be used with dataclass/attrs class and creates a subparser (got: {sub_parsers_field.type})"
+            )
+        sub_parser = subparsers.add_parser(
+            command, sub_parsers_args.name.lower(), **sub_parsers_args.parser_params
+        )
+        _make_parser(sub_parsers_args, sub_parser)
+
+
 def parse(cls: Type[T], args: Optional[Sequence[str]] = None, *, parser=None) -> T:
     """
     Parse command line arguments according to the fields of `cls` and populate it.
@@ -211,15 +316,75 @@ def parse(cls: Type[T], args: Optional[Sequence[str]] = None, *, parser=None) ->
     >>> parse(Args, ["--num", "1"])
     Args(is_flag=False, num=1)
     """
-    return cls(**vars(make_parser(cls, parser=parser).parse_args(args)))
+    result = vars(make_parser(cls, parser=parser).parse_args(args))
+    try:
+        command_dest = cls.__datargs_params__.sub_commands.get("dest", None)
+    except AttributeError:
+        pass
+    else:
+        if command_dest is not None and command_dest in result:
+            del result[command_dest]
+    return cls(**result)
 
 
-def argsclass(*args, **kwargs):
-    return dataclass(*args, **kwargs)
+def argsclass(
+    cls: type = None,
+    *args,
+    description: str = None,
+    parser_params: dict = None,
+    **kwargs,
+):
+    """
+    A wrapper around `dataclass` for passing `description` and other params (in `parser_params`)
+    to the `ArgumentParser` constructor.
+    """
+    # sub_commands_params has been disabled until a useful use case is found
+    datargs_kwargs = {
+        "description": description,
+        "parser_params": parser_params,
+        "sub_commands_params": {},
+    }
+    if cls is None:
+        # We're called with parens.
+        return partial(
+            make_class,
+            *args,
+            **datargs_kwargs,
+            **kwargs,
+        )
+
+    return make_class(
+        cls,
+        *args,
+        **datargs_kwargs,
+        **kwargs,
+    )
+
+
+def make_class(
+    cls,
+    description: str = None,
+    parser_params: dict = None,
+    sub_commands_params: dict = None,
+    *args,
+    **kwargs,
+):
+    try:
+        RecordClass.wrap_class(cls)
+    except NotARecordClass:
+        new_cls = dataclass(*args, **kwargs)(cls)
+    else:
+        new_cls = cls
+    new_cls.__datargs_params__ = DatargsParams(
+        parser={"description": description, **(parser_params or {})},
+        sub_commands=sub_commands_params or {},
+    )
+    return new_cls
 
 
 # noinspection PyShadowingBuiltins
 def arg(
+    positional=False,
     nargs=None,
     const=None,
     default=MISSING,
@@ -251,19 +416,26 @@ def arg(
       --invisible-arg MY_ARG    argument description
     """
     return dataclasses.field(
-        metadata=dict(
-            nargs=nargs,
-            choices=choices,
-            const=const,
-            help=help,
-            metavar=metavar,
-            aliases=aliases,
+        metadata=remove_dict_nones(
+            dict(
+                nargs=nargs,
+                choices=choices,
+                const=const,
+                help=help,
+                metavar=metavar,
+                aliases=aliases,
+                positional=positional,
+            )
         ),
         default=default,
         **kwargs,
     )
 
 
+def remove_dict_nones(dct: dict) -> dict:
+    return {key: value for key, value in dct.items() if value is not None}
+
+
 if __name__ == "__main__":
     import doctest
 
